import os

__version__ = "2.1.0"
API_ROOT = "https://v3.openstates.org"
DEFAULT_USER_AGENT = f"pyopenstates/{__version__}"
API_KEY_ENV_VAR = "OPENSTATES_API_KEY"
ENVIRON_API_KEY = os.environ.get("OPENSTATES_API_KEY")



import csv
import io
import pathlib
import requests
import tempfile
import zipfile
from enum import Enum

from .config import ENVIRON_API_KEY

TEMP_PATH = pathlib.Path(tempfile.gettempdir()) / "OS_ZIP_CACHE"


class FileType(Enum):
    """
    enum specifying the various types of files available from the CSV bulk data:

    - `Bills`
    - `Actions`
    - `Sources`
    - `Sponsorships`
    - `Versions`
    - `VersionLinks`
    - `Votes`
    - `VotePeople`
    - `VoteSources`
    - `Organizations`
    """

    Bills = "_bills.csv"
    Actions = "_bill_actions.csv"
    Sources = "_bill_sources.csv"
    Sponsorships = "_bill_sponsorships.csv"
    Versions = "_bill_versions.csv"
    VersionLinks = "_bill_version_links.csv"
    Votes = "_votes.csv"
    VotePeople = "_vote_people.csv"
    VoteSources = "_vote_sources.csv"
    Organizations = "_organizations.csv"


def _get_download_url(jurisdiction: str, session: str) -> str:
    url = f"https://v3.openstates.org/jurisdictions/{jurisdiction}?apikey={ENVIRON_API_KEY}&include=legislative_sessions"
    sessions = requests.get(url).json()["legislative_sessions"]
    for ses in sessions:
        if ses["identifier"] == session:
            break
    else:
        raise ValueError("invalid session")
    return ses["downloads"][0]["url"]


def _download_zip(url: str) -> pathlib.Path:
    filename = url.split("/")[-1]
    local_path = TEMP_PATH / filename
    TEMP_PATH.mkdir(parents=True, exist_ok=True)
    if not local_path.exists():
        with open(local_path, "wb") as f:
            f.write(requests.get(url).content)
    return local_path


def _load_session_data(state: str, session: str, file_type: FileType) -> str:
    url = _get_download_url(state, session)
    zip_path = _download_zip(url)
    with zipfile.ZipFile(zip_path) as zf:
        for filename in zf.namelist():
            if filename.endswith(file_type.value):
                break
        else:
            raise ValueError(f"no file of type {file_type} in {zip_path}")
        with zf.open(filename) as df:
            return df.read().decode()


def load_csv(state: str, session: str, file_type: FileType):
    """
    Returns an instantiated `csv.DictReader` to iterate over the requested file.
    """
    data = _load_session_data(state, session, file_type)
    return csv.DictReader(io.StringIO(data))


def load_merged_dataframe(state: str, session: str, which: FileType):
    """
    Returns a populated `pandas.DataFrame` with the requested content.

    `FileType.Actions`, `FileType.Sources`, `FileType.Versions`, `FileType.Sponsorships` will be merged against a `FileType.Bills` dataframe.

    `FileType.VersionLinks` will be  merged against both a `FileType.Versions` and `FileType.Bills` dataframe.

    `FileType.VotePeople` and `FileType.VoteSources` will be merged against a `FileType.Votes` dataframe.

    Other types will be returned as-is.
    """
    import pandas as pd

    other_df = pd.DataFrame(load_csv(state, session, which))

    if which in (
        FileType.Actions,
        FileType.Sources,
        FileType.Versions,
        FileType.Sponsorships,
    ):
        # these merge to Bills
        main_df = pd.DataFrame(load_csv(state, session, FileType.Bills))
        return main_df.merge(
            other_df,
            left_on="id",
            right_on="bill_id",
            how="left",
            suffixes=["_bill", ""],
        )
    elif which == FileType.VersionLinks:
        main_df = pd.DataFrame(load_csv(state, session, FileType.Bills))
        versions_df = pd.DataFrame(load_csv(state, session, FileType.Versions))
        main_df = main_df.merge(
            versions_df,
            left_on="id",
            right_on="bill_id",
            how="left",
            suffixes=["_bill", "_version"],
        )
        return main_df.merge(
            other_df,
            left_on="id_version",
            right_on="version_id",
            how="left",
            suffixes=["", "_link"],
        )
    elif which in (FileType.VotePeople, FileType.VoteSources):
        main_df = pd.DataFrame(load_csv(state, session, FileType.Votes))
        return main_df.merge(
            other_df,
            left_on="id",
            right_on="vote_event_id",
            how="left",
            suffixes=["_vote", ""],
        )
    else:
        return other_df



# -*- coding: utf-8 -*-

"""A Python client for the Open States API"""

import dateutil.parser
from requests import Session
from time import sleep

"""Copyright 2016 Sean Whalen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from .config import (  # noqa
    __version__,
    API_ROOT,
    DEFAULT_USER_AGENT,
    API_KEY_ENV_VAR,
    ENVIRON_API_KEY,
)

session = Session()
session.headers.update({"Accept": "application/json"})
session.headers.update({"User-Agent": DEFAULT_USER_AGENT})
if ENVIRON_API_KEY:
    session.headers.update({"X-Api-Key": ENVIRON_API_KEY})
else:
    print(f"Warning: No API Key found, set {API_KEY_ENV_VAR}")


class APIError(RuntimeError):
    """
    Raised when the Open States API returns an error
    """

    pass


class NotFound(APIError):
    """Raised when the API cannot find the requested object"""

    pass


def _make_params(**kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}


def _get(uri, params=None):
    """
    An internal method for making API calls and error handling easy and
    consistent

    Args:
        uri: API URI
        params: GET parameters

    Returns:
        JSON as a Python dictionary
    """

    def _convert_timestamps(result):
        """Converts a string timestamps from an api result API to a datetime"""
        if isinstance(result, dict):
            for key in result.keys():
                if key in (
                    "created_at",
                    "updated_at",
                    "latest_people_update",
                    "latest_bill_update",
                ):
                    try:
                        result[key] = dateutil.parser.parse(result[key])
                    except ValueError:
                        pass
                elif isinstance(result[key], dict):
                    result[key] = _convert_timestamps(result[key])
                elif isinstance(result[key], list):
                    result[key] = [_convert_timestamps(r) for r in result[key]]
        elif isinstance(result, list):
            result = [_convert_timestamps(r) for r in result]

        return result

    def _convert(result):
        """Convert results to standard Python data structures"""
        result = _convert_timestamps(result)
        return result

    url = f"{API_ROOT}/{uri}"
    for param in params.keys():
        if isinstance(params[param], list):
            params[param] = ",".join(params[param])
    response = session.get(url, params=params)
    if response.status_code != 200:
        if response.status_code == 404:
            raise NotFound(f"Not found: {response.url}")
        else:
            raise APIError(response.text)
    return _convert(response.json())


def set_user_agent(user_agent):
    """Appends a custom string to the default User-Agent string
    (e.g. ``pyopenstates/__version__ user_agent``)"""
    session.headers.update({"User-Agent": f"{DEFAULT_USER_AGENT} {user_agent}"})


def set_api_key(apikey):
    """Sets API key. Can also be set as OPENSTATES_API_KEY environment
    variable."""
    session.headers["X-Api-Key"] = apikey


def get_metadata(state=None, fields=None):
    """
        Returns a list of all states with data available, and basic metadata
        about their status. Can also get detailed metadata for a particular
        state.

    Args:
        state: The abbreviation of state to get detailed metadata on, or leave
        as None to get high-level metadata on all states.

        fields: An optional list of fields to return; returns all fields by
            default

    Returns:
       Dict: The requested :ref:`Metadata` as a dictionary
    """
    uri = "jurisdictions"
    params = dict()
    if state:
        uri += "/" + _jurisdiction_id(state)
        state_response = _get(uri, params=params)
        if fields is not None:
            return {k: state_response[k] for k in fields}
        else:
            return state_response
    else:
        params["page"] = "1"
        params["per_page"] = "52"
        return _get(uri, params=params)["results"]


def get_organizations(state):
    uri = "jurisdictions"
    uri += "/" + _jurisdiction_id(state)
    state_response = _get(uri, params={"include": "organizations"})
    return state_response["organizations"]


def search_bills(**kwargs):
    """
    Find bills matching a given set of filters

    Args:
        **kwargs: One or more search filters

    - ``state`` - Only return bills from a given state (e.g. ``nc``)
    - ``chamber`` - Only return bills matching the provided chamber
    (``upper`` or ``lower``)
    - ``bill_id`` - Only return bills with a given bill_id.
    - ``bill_id_in`` - Accepts a pipe (|) delimited list of bill ids.
    - ``q`` -  Only return bills matching the provided full text query.
    - ``search_window``- By default all bills are searched, but if a time
    window is desired the following options can be passed to
    ``search_window``:
        - ``search_window="all"`` - Default, include all sessions.
        - ``search_window="session"`` - Only bills from sessions within the
        current session.
        - ``search_window="session"`` - Only bills from the current session.
        - ``search_window="session:2009"`` - Only bills from the session named
        ``2009``.
        - ``search_window="term:2009-2011"`` - Only bills from the sessions in
        the ``2009-2011`` session.
    - ``updated_since`` - Only bills updated since a provided date (provided in
    ``YYYY-MM-DD`` format)
    - ``sponsor_id`` Only bills sponsored by a given legislator id (e.g.
    ``ILL000555``)
    - ``subject`` - Only bills categorized by Open States as belonging to this
    subject.
    - ``type`` Only bills of a given type (e.g. ``bill``, ``resolution``,
    etc.)

    You can specify sorting using the following ``sort`` keyword argument
    values:

    - ``first``
    - ``last``
    - ``signed``
    - ``passed_lower``
    - ``passed_upper``
    - ``updated_at``
    - ``created_at``

    Returns:
        A list of matching :ref:`Bill` dictionaries

    .. NOTE::
        This method returns just a subset (``state``, ``chamber``, ``session``,
        ``subjects``, ``type``, ``id``, ``bill_id``, ``title``, ``created_at``,
        ``updated_at``) of the bill fields by default.

        Use the ``fields`` parameter to specify a custom list of fields to
        return.
    """
    uri = "bills/"
    if "state" in kwargs.keys():
        kwargs["jurisdiction"] = _jurisdiction_id(kwargs["state"])

    if len(kwargs) > 0:
        kwargs["per_page"] = 20
        kwargs["page"] = 1
    results = []
    new_results = _get(uri, params=kwargs)["results"]
    while len(new_results) > 0:
        results += new_results
        kwargs["page"] += 1
        sleep(1)
        # When the search runs out of pages, the API returns not found
        try:
            new_results = _get(uri, params=kwargs)["results"]
        except NotFound:
            break

    return results


def get_bill(uid=None, state=None, session=None, bill_id=None, **kwargs):
    """
    Returns details of a specific bill Can be identified my the Open States
    unique bill id (uid), or by specifying the state, session, and
    legislative bill ID

    Args:
        uid: The Open States unique bill ID
        state: The postal code of the state
        session: The legislative session (see state metadata)
        bill_id: Yhe legislative bill ID (e.g. ``HR 42``)
        **kwargs: Optional keyword argument options, such as ``fields``,
        which specifies the fields to return

    Returns:
        The :ref:`Bill` details as a dictionary
    """
    if uid:
        if state or session or bill_id:
            raise ValueError(
                "Must specify an Open States bill (uid), or the "
                "state, session, and bill ID"
            )
        uid = _fix_id_string("ocd-bill/", uid)
        return _get(f"bills/{uid}", params=kwargs)
    else:
        if not state or not session or not bill_id:
            raise ValueError(
                "Must specify an Open States bill (uid), "
                "or the state, session, and bill ID"
            )
        return _get(f"bills/{state.lower()}/{session}/{bill_id}", params=kwargs)


def search_legislators(
    jurisdiction=None,
    name=None,
    id_=None,
    org_classification=None,
    district=None,
    include=None,
):
    """
    Search for legislators.

    Returns:
        A list of matching :ref:`Legislator` dictionaries

    """
    params = _make_params(
        jurisdiction=jurisdiction,
        name=name,
        id=id_,
        org_classification=org_classification,
        district=district,
        include=include,
    )
    return _get("people", params)["results"]


def get_legislator(leg_id):
    """
    Gets a legislator's details

    Args:
        leg_id: The Legislator's Open States ID
        fields: An optional custom list of fields to return

    Returns:
        The requested :ref:`Legislator` details as a dictionary
    """
    leg_id = _fix_id_string("ocd-person/", leg_id)
    return _get("people/", params={"id": [leg_id]})["results"][0]


def locate_legislators(lat, lng, fields=None):
    """
    Returns a list of legislators for the given latitude/longitude coordinates

    Args:
        lat: Latitude
        long: Longitude
        fields: An optional custom list of fields to return

    Returns:
        A list of matching :ref:`Legislator` dictionaries

    """
    return _get(
        "people.geo/", params=dict(lat=float(lat), lng=float(lng), fields=fields)
    )["results"]


def search_districts(state, chamber):
    """
    Search for districts

    Args:
        state: The state to search in
        chamber: the upper or lower legislative chamber
        fields: Optionally specify a custom list of fields to return

    Returns:
       A list of matching :ref:`District` dictionaries
    """
    if chamber:
        chamber = chamber.lower()
        if chamber not in ["upper", "lower"]:
            raise ValueError('Chamber must be "upper" or "lower"')
        organizations = get_organizations(state=state)
        for org in organizations:
            if org["classification"] == chamber:
                return org["districts"]


def _fix_id_string(prefix, id):
    if id.startswith(prefix):
        return id
    else:
        return prefix + id


def _jurisdiction_id(state):
    if state.startswith("ocd-jurisdiction/"):
        return state
    else:
        return f"ocd-jurisdiction/country:us/state:{state.lower()}/government"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyopenstates']

package_data = \
{'': ['*']}

install_requires = \
['python-dateutil>=2.8.2,<3.0.0', 'requests>=2.26.0,<3.0.0']

extras_require = \
{'pandas': ['pandas>=1.3.4,<2.0.0']}

setup_kwargs = {
    'name': 'pyopenstates',
    'version': '2.1.0',
    'description': 'A Python module for accessing the Open States API & bulk data.',
    'long_description': '# pyopenstates\n\nA Python client for the [Open States API v3](https://v3.openstates.org/docs).\n\nSource: [https://github.com/openstates/pyopenstates/](https://github.com/openstates/pyopenstates)\n\nDocumentation: [https://openstates.github.io/pyopenstates/](https://openstates.github.io/pyopenstates/)\n\nIssues: [https://github.com/openstates/pyopenstates/issues](https://github.com/openstates/pyopenstates/issues)\n\n**Note: This library was recently updated to support Open States API v3, documentation & coverage is a bit behind, but we wanted to get a release out.  Feel free to contribute issues and/or fixes.**\n\n[![PyPI badge](https://badge.fury.io/py/pyopenstates.svg)](https://badge.fury.io/py/pyopenstates)\n[![Test Python](https://github.com/openstates/pyopenstates/actions/workflows/main.yml/badge.svg)](https://github.com/openstates/pyopenstates/actions/workflows/main.yml)\n\n## Features\n\n- Compatible with Python 3.7+\n- Automatic conversion of string dates and timestamps to ``datetime`` objects.\n- Tested releases.\n- Set API Key via environment variable or in code.\n\n## API Keys\n\nTo use the Open States API you must [obtain an API Key](https://openstates.org/accounts/register/).\n\nOnce you have your key you can use it with this library by setting the ``OPENSTATES_API_KEY`` environment variable or calling ``pyopenstates.set_api_key``.\n\n## About Open States\n\nOpen States strives to improve civic engagement at the state level by providing data and tools regarding state legislatures. We aim to serve members of the public, activist groups, journalists, and researchers with better data on what is happening in their state capital, and to provide tools to reduce barriers to participation and increase engagement.\n\nOpen States aggregates legislative information from all 50 states, Washington, D.C., and Puerto Rico. This information is then standardized, cleaned, and published to the public via OpenStates.org, a powerful API, and bulk downloads. OpenStates.org enables individuals to find out who represents them, look up information on an important bill that’s been in the news, discover how their representatives are voting, or just stay current with what is happening in their state. Additionally, our API and bulk downloads see millions of hits every month from advocacy organizations, journalists, researchers, and many others.\n\nLegislative data is collected from official sources, linked at the bottom of relevant pages. In general bill & vote data is collected multiple times a day via our scrapers while legislator data is curated by our team & volunteers like you.\n\n',
    'author': 'James Turk',
    'author_email': 'james@openstates.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/openstates/pyopenstates',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
